local schema = require "model.schema"
local logger = require "logger"
local stringify = require "stringify"
local asset = require "model.asset"
local common_fun = require "model.common_fun"
local reward = require "model.reward"

local COINS_DUNGEONS = 3 -- 金币副本
local EXP_DUNGEONS = 4   -- 经验副本
local ICE_DUNGEONS = 5   -- 冰封祭坛

local MAX_DIFF = 6

local dungeons_list = {
    COINS_DUNGEONS,
    EXP_DUNGEONS,
    ICE_DUNGEONS,
}

local module_name = "daily_dungeons"
local module_id = MODULE_ID_DAILY_DUNGEONS
local _M = schema.new(module_name, {
    list = {
        --[[
            [type] = {
                type = type,
                diff = 1,
                num = 0,
            }
        ]]
    }
})

local REQUEST = {}
local CMD = {}
local MODULE = {}
local THIS = {}

local function func_ret(fname, character, args)
    local f = THIS[fname]
    if not f then
        logger.error("func_ret not fname:%s !!!", fname)
        return {errno = STD_ERR.COMMON_SYS_ERR}
    end
    local errno, ret = f(character, args)
    if errno ~= 0 then
        return {errno = errno}
    end
    ret = ret or {}
    ret.errno = 0
    return ret
end

function MODULE.list_request_interests() return REQUEST end
function MODULE.list_command_interests() return CMD end

-- TODO: 解析/升级模块数据 在这里把数据初始化好
function MODULE.parse(character)
    local d = _M.load(character)
    d.list = d.list or {}
    local change = false
    for _, dtype in ipairs(dungeons_list) do
        if not d.list[dtype] then
            local change = true
            d.list[dtype] = {
                type = dtype, 
                diff = 0,
                num = 0,
            }
        end
    end
    if change then
        _M.persist(character)
    end
end

-- TODO: 侦听事件
function MODULE.monitor(character)
    local d = _M.assert_get(character)
    character.monitor("daily_refresh", function()
        local change = false
        for _, data in pairs(d.list) do
            if data.num > 0 then
                data.num = 0
                change = true
            end
        end
        if change then
            _M.persist(character)
        end
    end)
end

-- TODO: 类似泰利的 prepare 接口
function MODULE.launch(character)
    local d = _M.assert_get(character)
end

-- TODO: 与客户端同步数据
function MODULE.ready(character)
    local d = _M.assert_get(character)
    logger.test("%s:ready, %s", module_name, stringify(d or {}))
end

-- TODO: 玩家下线时的处理
function MODULE.saybye(character)

end

local function create_client_data(data)
    return {
        type = data.type,
        diff = data.diff,
        num = data.num
    }
end

function THIS.daily_dungeons_data(character, args)
    local d = _M.assert_get(character) or {}
    local list = {}
    for _, data in pairs(d.list) do
        table.insert(list, create_client_data(data))
    end
    return 0, {list = list}
end

function THIS.daily_dungeons_start(character, args)
    if not args.type or not args.diff then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local d = _M.assert_get(character) or {}
    local data = d.list[args.type]
    if not data then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local max_time = 0
    local data_conf = asset.DataConfig_proto[45]
    if args.type == COINS_DUNGEONS then
        max_time = data_conf.data1
    elseif args.type == EXP_DUNGEONS then
        max_time = data_conf.data2
    elseif args.type == ICE_DUNGEONS then
        max_time = data_conf.data3
    end

    if data.num + 1 > max_time then
        return STD_ERR.DAILY_DUNGEONS_BATTLE_TIMES_LIMTE -- 挑战次数不足
    end

    if args.diff > data.diff+1 or args.diff > MAX_DIFF then
        return STD_ERR.DAILY_DUNGEONS_CANT_PASS -- 无法挑战的难度
    end

    local daily_conf
    for _, conf in pairs(asset.DailyConfig_proto) do
        if conf.diff == args.diff and conf.destype == args.type then
            daily_conf = conf 
            break
        end
    end

    if not daily_conf then
        return STD_ERR.COMMON_CONF_ERR -- 配置异常
    end

    local errno, award_list = common_fun.get_award(daily_conf.reward)
    if errno ~= 0 then
        return errno
    end

    if not next(award_list or {}) then
        return STD_ERR.COMMON_CONF_ERR -- 配置异常
    end

    data.num = data.num + 1
    d.battle_id = daily_conf.ID
    d.award_list = award_list
    _M.persist(character)
    local temp = {}
    for _, v in ipairs(award_list or {}) do
        table.insert(temp, {id = v.id, num = v.num})
    end
    return 0, {award_list = temp}
end

function THIS.daily_dungeons_end(character, args)
    local d = _M.assert_get(character) or {}
    local max_list = {}
    local all_num = 0

    if not d.battle_id then
        return STD_ERR.DAILY_DUNGEONS_NOT_BATTLE -- 没有进行中的战斗
    end

    local daily_conf = asset.DailyConfig_proto[d.battle_id]
    if not daily_conf then
        return STD_ERR.COMMON_SYS_ERR -- 系统异常
    end

    local data = d.list[daily_conf.destype]
    if not data then
        return STD_ERR.COMMON_SYS_ERR -- 系统异常
    end

    for _, id in ipairs(daily_conf.generate) do
        local challenge_conf = asset.ChallengeConfig_proto[id]
        if challenge_conf then
            all_num = all_num + #challenge_conf.roles
        end
        for _, monster_id in ipairs(challenge_conf.roles) do
            local monster_conf = asset.MonsterConfig_proto[monster_id]
            if monster_conf then
                max_list[monster_conf.type] = (max_list[monster_conf.type] or 0) + 1
            end
        end 
    end

    local kill_num = all_num
    local kill_elite = max_list[MONESTER_TYPE_BOSS] or 0
    local kill_boss = max_list[MONESTER_TYPE_BOSS] or 0
    -- if not args.win then
    local monster_num = args.monster_num or 0
    local elite_num = args.elite_num or 0
    local boss_num = args.boss_num or 0
    kill_num = math.min(monster_num + elite_num + boss_num, all_num)
    kill_elite = math.min(elite_num, kill_elite)
    kill_boss = math.min(boss_num, kill_boss)
    -- end
    local award_list = d.award_list or {}
    award_list = common_fun.award_multiplier(award_list, kill_num)
    
    if args.win and daily_conf.clearance > 0 then
        local errno, temp_list = common_fun.get_award(daily_conf.clearance)
        if errno ~= 0 then
            logger.warn("get_award err, awardid:%d", daily_conf.clearance)
        end
        if next(temp_list or {}) then
            award_list = common_fun.merge2list(award_list, temp_list)
        end
    end

    d.battle_id = nil
    d.award_list = nil
    
    if args.win and daily_conf.diff > data.diff then
        data.diff = daily_conf.diff
    end

    if next(award_list or {}) then
        -- 发奖励
        local desc = {
            module="daily_dungeons",
            brief="日常副本",
            context= "日常副本, type:%d, diff:%d", daily_conf.destype, daily_conf.diff,
            mailgen={subject=2, body=""},
            notify={
                flags="daily_dungeons_end"
            },
            detail=award_list
        }
        reward(character, desc)
    end
    _M.persist(character)

    if kill_num > 0 then
        character.dispatch("kill_monster", kill_num)
    end
    if kill_boss > 0 then
        character.dispatch("kill_boss", kill_boss)
    end
    character.dispatch("daily_dungeons_pass", daily_conf.destype, daily_conf.diff)
    return 0, {data = create_client_data(data)}
end

function THIS.daily_dungeons_sweep(character, args)
    if not args.type or not args.diff then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local d = _M.assert_get(character)
    local data = d.list[args.type]
    if not data then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local max_time = 0
    local data_conf = asset.DataConfig_proto[45]
    if args.type == COINS_DUNGEONS then
        max_time = data_conf.data1
    elseif args.type == EXP_DUNGEONS then
        max_time = data_conf.data2
    elseif args.type == ICE_DUNGEONS then
        max_time = data_conf.data3
    end

    if data.num + 1 > max_time then
        return STD_ERR.DAILY_DUNGEONS_BATTLE_TIMES_LIMTE -- 挑战次数不足
    end

    if data.diff < args.diff then
        return STD_ERR.DAILY_DUNGEONS_CANT_PASS -- 未通关
    end

    local all_num = 0
    local daily_conf
    for _, conf in pairs(asset.DailyConfig_proto) do
        if conf.diff == args.diff and conf.destype == args.type then
            daily_conf = conf 
            break
        end
    end

    if not daily_conf then
        return STD_ERR.COMMON_SYS_ERR -- 系统异常
    end

    for _, id in ipairs(daily_conf.generate) do
        local challenge_conf = asset.ChallengeConfig_proto[id]
        if challenge_conf then
            all_num = all_num + #challenge_conf.roles
        end
    end

    local errno, award_list = common_fun.get_award(daily_conf.reward)
    if errno ~=0 then
        return errno
    end

    award_list = common_fun.award_multiplier(award_list, all_num)
    
    if  daily_conf.clearance > 0 then
        local errno, temp_list = common_fun.get_award(daily_conf.clearance)
        if errno ~= 0 then
            logger.warn("get_award err, awardid:%d", daily_conf.clearance)
            return errno
        end
        if next(temp_list or {}) then
            award_list = common_fun.merge2list(award_list, temp_list)
        end
    end

    if next(award_list or {}) then
        -- 发奖励
        local desc = {
            module="daily_dungeons",
            brief="日常副本扫荡",
            context= "日常副本扫荡, type:%d, diff:%d", daily_conf.destype, daily_conf.diff,
            mailgen={subject=2, body=""},
            notify={
                flags="daily_dungeons_sweep"
            },
            detail=award_list
        }
        reward(character, desc)
    end
    data.num = data.num + 1

    _M.persist(character)
    character.dispatch("daily_dungeons_sweep", daily_conf.destype, daily_conf.diff)
    character.dispatch("watch_ad")
    return 0, {data = create_client_data(data)}
end

function REQUEST.daily_dungeons_data(character, args)
    return func_ret("daily_dungeons_data", character, args)
end

function REQUEST.daily_dungeons_start(character, args)
    return func_ret("daily_dungeons_start", character, args)
end

function REQUEST.daily_dungeons_end(character, args)
    return func_ret("daily_dungeons_end", character, args)
end

function REQUEST.daily_dungeons_sweep(character, args)
    return func_ret("daily_dungeons_sweep", character, args)
end

-- 红点
function MODULE.get_red_point(character)
    local d = _M.assert_get(character) or {}
    local data_conf = asset.DataConfig_proto[45]
    local max_times = 0
    for dtype, data in pairs(d.list) do
        if dtype == COINS_DUNGEONS then
            max_times = data_conf.data1
        elseif dtype == EXP_DUNGEONS then
            max_times = data_conf.data2
        elseif dtype == ICE_DUNGEONS then
            max_times = data_conf.data3
        end
        if data.num < max_times then
            return true, module_id
        end
    end
end


return MODULE