local schema = require "model.schema"
local logger = require "logger"
local stringify = require "stringify"
local asset = require "model.asset"
local util = require "util"
local reward = require "model.reward"
local common_fun = require "model.common_fun"
local payment = require "model.payment"
local skynet = require "skynet"
local relic_manual = skynet.localname(".relic_manual")

local MAX_LAYER = 3 -- 最大层数
local MAX_POS = 11 -- 每层最大位置
local MAX_DIFF = 6  -- 最高难度

local STATE_OVER = 0 -- 结束
local STATE_BUFF = 1 -- 需选择buff

-- 异常时默认挑战关卡
local DEFAULT_BATTLEID1 = 1     -- 普通挑战
local DEFAULT_BATTLEID2 = 2     -- 精英战斗
local DEFAULT_BATTLEID3 = 34    -- 驻军战斗
local DEFAULT_BATTLEID4 = 25    -- boss战斗

local BUFF_SINGLE = 0 -- 单次出现的buff
local BUFF_MULTIPLE = 1 -- 多次出现的buff

local EVENT_BATTLE = 2   -- 战斗
local EVENT_HOT_SPRING = 3 -- 温泉回体力
local EVENT_SELECT_BUFF = 4 -- 选buff

local ADV_BATTLE1 = 1 -- 普通战斗
local ADV_BATTLE2 = 2 -- 精英战斗
local ADV_BATTLE3 = 3 -- 驻军战斗
local ADV_HOT_SPRING = 4 -- 温泉
local ADV_AWARD1 = 5 -- 钻石奖励
local ADV_AWARD2 = 6 -- 随机奖励
local ADV_BOSS = 7 -- boss战斗

local PATH_TABLE = {
    [0] = {1},
    [1] = {2, 3},
    [2] = {4, 5},
    [3] = {5, 6},
    [4] = {7},
    [5] = {7, 8},
    [6] = {8},
    [7] = {9},
    [8] = {9, 10},
    [9] = {11},
    [10] = {11},
    [11] = {0}
}

local module_name = "relic"
local moduld_id = MODULE_ID_RELIC
local _M = schema.new(module_name, {
    max_diff = 0,   -- 历史最高难度
    cur_diff = 0,   -- 当前难度
    buff_list = {},  -- 玩家buff
    select_buff = {}, -- 待选的buff
    pos_list = {},  -- 路径数据
    pos_battle_info = {}, --每个格子的战斗数据
    layer = 0,      -- 在第几层
    pos = 0,        -- 位置
    cost = 0,       -- 当前费用
    state = 0,      -- 状态 0结束, 1选buff
    reset_time = 0, -- 重置时间


    battle_list1 = {}, -- 普通困难关卡战斗id
    battle_list2 = {}, -- 驻军关卡战斗id
    battle_list3 = {}, -- boss关卡战斗id
})

local REQUEST = {}
local CMD = {}
local MODULE = {}
local THIS = {}

local function send_evnet_msg(character, msg)
    character.send("relic_event_nty", msg)
end

local function send_award(character, award_list, layer, pos, flags)
    flags = flags or "relic_next_pos"
    if not award_list or not next(award_list) then
        return 
    end
    local desc = {
        module="relic",
        brief="遗迹奖励",
        context= string.format("遗迹奖励, layer: %d, pos:%d", layer, pos),
        mailgen={subject=2, body=""},
        notify={
            flags=flags
        },
        detail=award_list
    }
    reward(character, desc)
end

local function create_award(conf)
    local award_list = {}
    for _, award_id in ipairs(conf.random) do
        if award_id <= 0 then
            break
        end
        local _, temp = common_fun.get_award(award_id)
        if temp and next(temp) then
            if not next(award_list) then
                award_list = temp
            else
                award_list = common_fun.merge2list(award_list, temp)                    
            end
        end
    end
    if conf.reward and conf.reward[1] > 0 then
        for i = 1, #conf.reward, 2 do
            local item_id = conf.reward[i]
            local item_num = conf.reward[i+1]
            if item_id and item_num then
                table.insert(award_list, {id = item_id, num = item_num})
            end
        end
    end
    return award_list
end

local function func_ret(fname, character, args)
    local f = THIS[fname]
    if not f then
        logger.error("func_ret not fname:%s !!!", fname)
        return {errno = STD_ERR.COMMON_SYS_ERR}
    end
    local errno, ret = f(character, args)
    if errno ~= 0 then
        return {errno = errno}
    end
    ret = ret or {}
    ret.errno = 0
    return ret
end

local function get_relic_conf_list(diff, layer)
    local relic_conf_list = {}
    for _, conf in pairs(asset.RelicConfig_proto) do
        if conf.difficult == diff and conf.layer == layer then
            relic_conf_list[conf.pos] = conf
        end
    end
    return relic_conf_list
end

local function create_relic_battle_info(character)
    local d = _M.assert_get(character) or {}
    local relic_conf_list = get_relic_conf_list(d.cur_diff, d.layer)
    d.pos_battle_info = {}
    for _, conf in ipairs(relic_conf_list) do
        local battle_list
        if conf.type == ADV_BATTLE1 or conf.type == ADV_BATTLE2 then
            battle_list = d.battle_list1
        elseif conf.type == ADV_BATTLE3 then
            battle_list = d.battle_list2
        elseif conf.type == ADV_BOSS then
            battle_list = d.battle_list3
        end
        if battle_list then
            local size = #battle_list
            if size > 0 then
                table.insert(d.pos_battle_info, {
                    pos = conf.pos,
                    type = conf.type,
                    id = battle_list[1]
                })
            end
            if size > 1 then
                table.remove(battle_list, 1)
            end
        end
    end
end

local function relic_reset(character, diff)
    local diff = diff or 1
    if diff > MAX_DIFF then
        diff = MAX_DIFF
    end
    local d = _M.assert_get(character)
    d.cur_diff = diff
    d.buff_list = {}
    d.select_buff = {}
    d.pos_list = {0}
    d.pos_battle_info = {}
    d.layer = 1
    d.pos = 0
    local data_conf = asset.DataConfig_proto[24]
    d.cost = data_conf.data1
    d.state = STATE_OVER
    local activitytime = skynet.localname(".activitytime")
    local open_time = util.today(skynet.call(activitytime, "lua", "query"))
    data_conf = asset.DataConfig_proto[22]
    local today = util.today()
    local reset_time = nil
    if open_time then
        local start_time = util.today(open_time+12*HOUR_SEC)
        if today >= start_time and  data_conf.data1 > 0 then
            local num = math.floor((today - start_time)/DAY_SEC/data_conf.data1)+1
            reset_time = start_time + num * data_conf.data1 * DAY_SEC
        end
    end
    if not reset_time then
        reset_time = today + data_conf.data1 * DAY_SEC
    end
    d.reset_time = reset_time

    local temp1 = {}
    local temp2 = {}
    local temp3 = {}
    
    for i, conf in pairs(asset.ReliclevelsConfig_proto) do
        if conf.grouptype == 1 then
            table.insert(temp1, i)
        elseif conf.grouptype == 3 then
            table.insert(temp2, i)
        elseif conf.grouptype == 2 then
            table.insert(temp3, i)
        end
    end
    d.battle_list1 = common_fun.table_rand(temp1)
    d.battle_list2 = common_fun.table_rand(temp2)
    d.battle_list3 = common_fun.table_rand(temp3)
    create_relic_battle_info(character)
    _M.persist(character)
end



local event_create = {}
event_create[EVENT_BATTLE] = function(id, battle_type)
    return {
        type = EVENT_BATTLE,
        battle_id = id,
        battle_type = battle_type,
    }
end

event_create[EVENT_HOT_SPRING] = function()
    return {
        type = EVENT_HOT_SPRING,
    }
end

event_create[EVENT_SELECT_BUFF] = function(list)
    return {
        type = EVENT_SELECT_BUFF,
        select_buff = list
    }
end

local event_doing = {}
event_doing[EVENT_BATTLE] = function(character, args)
    local ctx = _M.assert_runtime(character)
    local event_list = ctx.event_list
    if args.win then
        if not args.cost then
            return STD_ERR.COMMON_PARM_ERR -- 参数异常
        end
        if args.cost < 0 then
            return STD_ERR.COMMON_PARM_ERR -- 参数异常
        end
        local d = _M.assert_get(character) or {}

        d.cost = args.cost + (event_list.add_cost or 0)
        d.pos = assert(event_list.pos)
        table.insert(d.pos_list, d.pos)
        d.state = assert(event_list.state)
        if d.layer == MAX_LAYER and d.pos == MAX_POS then
            d.state = STATE_OVER
            if  d.max_diff < d.cur_diff then
                d.max_diff = d.cur_diff
                character.dispatch("relic_diff_pass", d.max_diff)
            end
        end
        _M.persist(character)
        send_award(character, event_list.award_list, d.layer, d.pos, "relic_event_select")
        THIS.relic_data_nty(character, event_list.add_cost)
        character.dispatch("relic_battle")
        ctx.event_list = {}
    else
        ctx.event_list = {}
    end
    return 0
end

event_doing[EVENT_HOT_SPRING] = function(character, args)
    if not args.select then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local add_cost = 0
    local pay_list = {}
    local data_conf = asset.DataConfig_proto[24]
    if args.select == 1 then
        add_cost = data_conf.data2
    elseif args.select == 2 then
        add_cost = data_conf.data3
        table.insert(pay_list, {id = CURRENCY_ID_DIA, num = data_conf.data4})
    end
    if next(pay_list) then
        local receipt = {
            module="relic",
            brief="遗迹",
            context="遗迹恢复体力",
            notify={
                flags="relic.relic_event_select"
            },
            detail=pay_list
        }
        -- 检查消耗
        local errno = payment(character, receipt)
        if errno ~= 0 then
            return errno
        end
    end


    local ctx = _M.assert_runtime(character)
    local event_list = ctx.event_list
    local d = _M.assert_get(character)
    d.pos = assert(event_list.pos)
    table.insert(d.pos_list, d.pos)
    d.state = assert(event_list.state)
    d.cost = d.cost + add_cost
    _M.persist(character)
    THIS.relic_data_nty(character)
    ctx.event_list = {}
    return 0
end

event_doing[EVENT_SELECT_BUFF] = function(character, args)
    if not args.select then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local d = _M.assert_get(character)
    local ctx = _M.assert_runtime(character)
    if d.state ~= STATE_BUFF then
        return STD_ERR.RELIC_CANT_SELECT_BUF -- 无法选择buff
    end

    local buff_id = d.select_buff[args.select]
    if not buff_id then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end
    table.insert(d.buff_list, buff_id)
    d.select_buff = {}
    d.state = STATE_OVER
    _M.persist(character)
    THIS.relic_data_nty(character)
    ctx.event_list = {}
    return 0
end



local func_adv = {}
func_adv[ADV_BATTLE1] = function(character, conf)
    local d = _M.assert_get(character)
    local list = {
        type = EVENT_BATTLE,
        pos = conf.pos,
        state = STATE_BUFF,
        award_list = create_award(conf),
        add_cost = conf.Ereward or 0
    }
    local battle_id = DEFAULT_BATTLEID1
    for _, v in ipairs(d.pos_battle_info) do
        if v.pos == conf.pos then
            battle_id = v.id        
            break
        end
    end
    local msg =  event_create[EVENT_BATTLE](battle_id, ADV_BATTLE1)
    send_evnet_msg(character, msg)
    return list
end

func_adv[ADV_BATTLE2] = function(character, conf)
    local d = _M.assert_get(character)
    local list = {
        type = EVENT_BATTLE,
        pos = conf.pos,
        state = STATE_BUFF,
        award_list = create_award(conf),
        add_cost = conf.Ereward or 0
    }
    local battle_id = DEFAULT_BATTLEID2
    for _, v in ipairs(d.pos_battle_info) do
        if v.pos == conf.pos then
            battle_id = v.id        
            break
        end
    end
    local msg =  event_create[EVENT_BATTLE](battle_id, ADV_BATTLE2)
    send_evnet_msg(character, msg)
    return list
end

func_adv[ADV_BATTLE3] = function(character, conf)
    local d = _M.assert_get(character)
    local list = {
        type = EVENT_BATTLE,
        pos = conf.pos,
        state = STATE_OVER,
        award_list = create_award(conf),
        add_cost = conf.Ereward or 0
    }
    local battle_id = DEFAULT_BATTLEID3
    for _, v in ipairs(d.pos_battle_info) do
        if v.pos == conf.pos then
            battle_id = v.id        
            break
        end
    end
    local msg = event_create[EVENT_BATTLE](battle_id, ADV_BATTLE3)
    send_evnet_msg(character, msg)
    return list
end

func_adv[ADV_HOT_SPRING] = function(character, conf)
    local list = {
        type = EVENT_HOT_SPRING,
        pos = conf.pos,
        state = STATE_OVER,
    }
    local msg = event_create[EVENT_HOT_SPRING]()
    send_evnet_msg(character, msg)
    return list
end

func_adv[ADV_AWARD1] = function(character, conf)
    local award_list = create_award(conf)
    local d = _M.assert_get(character)
    d.pos = conf.pos
    table.insert(d.pos_list, d.pos)
    d.state = STATE_OVER
    _M.persist(character)
    send_award(character, award_list, d.layer, d.pos)
    THIS.relic_data_nty(character)
end

func_adv[ADV_AWARD2] = function(character, conf)
    local award_list = create_award(conf)
    local d = _M.assert_get(character)
    d.pos = conf.pos
    table.insert(d.pos_list, d.pos)
    d.state = STATE_OVER
    _M.persist(character)
    send_award(character, award_list, d.layer, d.pos)
    THIS.relic_data_nty(character)
end

func_adv[ADV_BOSS] = function(character, conf)
    local d = _M.assert_get(character)
    local list = {
        type = EVENT_BATTLE,
        pos = conf.pos,
        state = STATE_BUFF,
        award_list = create_award(conf),
    }
    local battle_id = DEFAULT_BATTLEID4
    for _, v in ipairs(d.pos_battle_info) do
        if v.pos == conf.pos then
            battle_id = v.id        
            break
        end
    end
    local msg =  event_create[EVENT_BATTLE](battle_id, ADV_BOSS)
    send_evnet_msg(character, msg)
    return list
end

function MODULE.list_request_interests() return REQUEST end
function MODULE.list_command_interests() return CMD end

-- TODO: 解析/升级模块数据 在这里把数据初始化好
function MODULE.parse(character)
    local d = _M.load(character)
    if not d.cur_diff or d.cur_diff == 0 then
        relic_reset(character)
    end
    if not d.pos_battle_info then
        create_relic_battle_info(character)
    end
end

-- TODO: 侦听事件
function MODULE.monitor(character)

end

-- TODO: 类似泰利的 prepare 接口
function MODULE.launch(character)
    local d = _M.assert_get(character)
    local u = _M.assert_runtime(character)
end

-- TODO: 与客户端同步数据?
function MODULE.ready(character)
    local d = _M.assert_get(character)
    logger.test("%s:ready, %s", module_name, stringify(d or {}))
end

-- TODO: 玩家下线时的处理
function MODULE.saybye(character)
end




function THIS.relic_event_select(character, args)
    local ctx = _M.assert_runtime(character)
    if not ctx.event_list or not next(ctx.event_list) then
        return STD_ERR.RELIC_NOT_EVENT -- 没有事件
    end
    if ctx.event_list.type ~= args.type then
        return STD_ERR.RELIC_NOT_EVENT -- 没找到对应事件
    end
    return event_doing[args.type](character, args)
end

function THIS.relic_reset(character, args)
    local d = _M.assert_get(character)
    local now = os.time()
    local diff  = args.difficulty
    if diff <= 0 and diff > MAX_DIFF or diff > d.max_diff + 1 then
        return STD_ERR.RELIC_ERR_DIFF -- 无法挑战更高难度
    end

    if d.reset_time and d.reset_time > now then
        return STD_ERR.RELIC_CANT_RESET -- 未到重置时间
    end
    relic_reset(character, diff)
    THIS.relic_data_nty(character)
    return 0
end

function THIS.relic_select_buff(character, args)
    local d = _M.assert_get(character)
    if not d.select_buff or not next(d.select_buff) then
        local relic_conf 
        for _, v in pairs(asset.RelicConfig_proto) do
            if v.layer == d.layer and v.pos == d.pos then
                relic_conf = v
                break
            end
        end
        if not relic_conf then
            return 4 -- 配置异常
        end

        local cant_list = {}
        for _, id in ipairs(d.buff_list) do
            local conf = asset.RunecardConfig_proto[id]
            if conf and conf.eliminate == BUFF_SINGLE then 
                cant_list[id] = 1
            end
        end

        local func = function(list)
            local all = 0
            local blist = {}
            for i=1, #list, 2 do
                local buff_id = list[i]
                local num = list[i+1]
                local conf = asset.RunecardConfig_proto[buff_id]
                if not(cant_list[buff_id]) and conf and num then
                    table.insert(blist, {
                        id = buff_id,
                        num = num,
                        type = conf.eliminate,
                    })
                    all = all + num
                end
            end
            if all <= 0 then
                return 0
            end
            local randnum = 1
            if all > 1 then
                randnum = util.rand(1, all)
            end
            for _, v in ipairs(blist) do
                if v.num >= randnum then
                    if v.type == BUFF_SINGLE then
                        cant_list[v.id] = 1
                    end                        
                    return v.id
                else
                    randnum = randnum - v.num
                end
            end
            return 0
        end
        local buff_id1 = func(relic_conf.runetA)
        if buff_id1 == 0 then
            return 4 -- 配置异常
        end
        local buff_id2 = func(relic_conf.runetB)
        if buff_id2 == 0 then
            return 4 -- 配置异常
        end
        local buff_id3 = func(relic_conf.runetC)
        if buff_id3 == 0 then
            return 4 -- 配置异常
        end
        d.select_buff = {buff_id1, buff_id2, buff_id3}
        _M.persist(character)
    end

    if d.state ~= STATE_BUFF then
        return STD_ERR.RELIC_CANT_SELECT_BUF -- 无法选择buff
    end

    if not next(d.select_buff) then
        return STD_ERR.COMMON_SYS_ERR -- 系统异常
    end
    local msg = event_create[EVENT_SELECT_BUFF](d.select_buff)
    send_evnet_msg(character, msg)
    local ctx = _M.assert_runtime(character)
    ctx.event_list = {type = EVENT_SELECT_BUFF}
    return 0
end

function THIS.relic_next_layer(character, args)
    local d = _M.assert_get(character)
    if d.layer < MAX_LAYER and d.pos >= MAX_POS and d.state == STATE_OVER then
        d.layer = d.layer + 1
        d.pos = 0
        d.pos_list = {d.pos}
        d.state = STATE_OVER
        create_relic_battle_info(character)
        _M.persist(character)
        THIS.relic_data_nty(character)
    else
        return 1 -- 参数异常
    end
    return 0
end

function THIS.relic_next_pos(character, args)
    local pos = args.pos or 0
    local d = _M.assert_get(character) or {}
    if d.state ~= STATE_OVER then
        return STD_ERR.RELIC_UNRESOLVED_EVENTS -- 有未处理的事件
    end

    if pos > MAX_POS then
        return STD_ERR.COMMON_PARM_ERR -- 参数异常
    end

    local pos_list = PATH_TABLE[d.pos]
    local can_enter = false
    for _, v in ipairs(pos_list) do
        if v == pos then
            can_enter = true
            break
        end
    end
    if not can_enter then
        return STD_ERR.RELIC_CANT_ENTER_NEXT_POS -- 无法通行
    end
    local relic_conf_list = get_relic_conf_list(d.cur_diff, d.layer) or {}
    local relic_conf = relic_conf_list[pos]
    if not relic_conf then
        return STD_ERR.COMMON_CONF_ERR -- 配置异常
    end
    local ctx = _M.assert_runtime(character)
    ctx.event_list = func_adv[relic_conf.type](character, relic_conf)
    return 0
end

function THIS.relic_data(character, args)
    THIS.relic_data_nty(character)
    return 0
end

function THIS.relic_data_nty(character, add_cost)
    local d = _M.assert_get(character) or {}
    local battle_info = {}
    for k, v in ipairs(d.pos_battle_info) do
        if v.type == ADV_BATTLE3 then
            table.insert(battle_info, v.id)
        end
    end

    local ret = {
        max_diff = d.max_diff,
        cur_diff = d.cur_diff,
        buff_list = d.buff_list or {},
        layer = d.layer,
        pos = d.pos,
        pos_list = d.pos_list,
        cost = d.cost,
        state = d.state,
        reset_time = d.reset_time or 0,
        battle_info = battle_info,
        add_cost = add_cost or 0,
    }
    character.send("relic_data_nty", ret)
end

function REQUEST.relic_event_select(character, args)
    return func_ret("relic_event_select", character, args)
end

function REQUEST.relic_select_buff(character, args)
    return func_ret("relic_select_buff", character, args)
end

function REQUEST.relic_next_layer(character, args)
    return func_ret("relic_next_layer", character, args)
end

function REQUEST.relic_next_pos(character, args)
    return func_ret("relic_next_pos", character, args)
end

function REQUEST.relic_data(character, args)
    return func_ret("relic_data", character, args)
end

function REQUEST.relic_reset(character, args)
    return func_ret("relic_reset", character, args)
end

-- GM增加玩家费用
function MODULE.gm_add_cost(character, num)
    local d = _M.assert_get(character)
    d.cost = d.cost + num
    _M.persist(character)
end

function MODULE.get_red_point(character)
    local d = _M.assert_get(character) or {}
    local now = os.time()
    if d.reset_time and now > d.reset_time then
        return true, moduld_id
    end

    -- if d.cost > 3 and d.layer ~= MAX_LAYER and d.pos ~ MAX_POS then
    --     return true, moduld_id
    -- end
    
    local model =  require "model.manual.relic_manual"
    local red = model.get_red_point(character)
    if red then
        return red , moduld_id
    end
end


return MODULE